/*
 * This file is part of SionTower.
 *
 * 
 * David Saltares Márquez (C) 2011
 * <david.saltares@gmail.com>
 *
 * 
 * SionTower examples are free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License ad
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) ant later version.
 *
 * SionTower examples are distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SionTower examples.  If not, see <http://www.gnu.org/licenses/>.
 */



/**
 *  @file collisionManager.cpp
 *  @author David Saltares Márquez
 *  @date 03-02-2011
 */


#include <iostream>

#include "collisionManager.h"

using std::cout;
using std::endl;

template<> CollisionManager* Ogre::Singleton<CollisionManager>::ms_Singleton = 0;

CollisionManager::CollisionManager() {
    cout << "CollisionManager::ColisionManager()" << endl;

    // Registramos los tests
    Shape::configureCollisionDispatching();

    // Creamos la esferas de partición
    _sphereA = new Sphere("spherePartitionA", Ogre::Vector3(0, 0, 0), 10.0f);
    _sphereB = new Sphere("spherePartitionB", Ogre::Vector3(0, 0, 0), 10.0f);
}

CollisionManager::~CollisionManager() {
    cout << "CollisionManager::~ColisionManager()" << endl;
}

CollisionManager& CollisionManager::getSingleton() {
    assert(ms_Singleton);
    return (*ms_Singleton);
}

CollisionManager* CollisionManager::getSingletonPtr() {
    assert(ms_Singleton);
    return ms_Singleton;
}

void CollisionManager::addBody(Body* body) {
    _bodies.push_back(body);
}

bool CollisionManager::removeBody(Body* body) {
    std::list<Body*>::iterator i;

    for (i = _bodies.begin(); i != _bodies.end(); ++i){
        if (*i == body){
            _bodies.erase(i);
            return true;
        }
    }

    return false;
}

void CollisionManager::removeAllBodies() {
    _bodies.clear();
}

void CollisionManager::addCollisionCallback(int typeA, int typeB, CollisionCallback c) {
    _collisionCallbackTable[typeA][typeB] = c;
    _collisionCallbackTable[typeB][typeA] = c;
}

bool CollisionManager::removeCollisionCallback(int typeA, int typeB) {
    CollisionCallbackTable::iterator i;
    size_t numErased;

    // Si existe una entrada para A, eliminamos su relación con B
    i = _collisionCallbackTable.find(typeA);

    if (i != _collisionCallbackTable.end())
        numErased = i->second.erase(typeB);

    // Si existe una entrada para B, eliminamos su relacion con A
    i = _collisionCallbackTable.find(typeB);

    if (i != _collisionCallbackTable.end())
        i->second.erase(typeA);

    return numErased > 0;
}

void CollisionManager::removeAllCollisionCallbacks() {
    _collisionCallbackTable.clear();
}

void CollisionManager::checkCollisions() {
    // Cruzamos los Bodies y comprobamos:
    // - Existe un collisionCallback para su tipo
    // - Están a una distancia razonable (sphere test)
    // - Test de colisión profundo
    // - Llamada al callback

    std::list<Body*>::iterator i;
    std::list<Body*>::iterator j;
    CollisionCallback collisionCallback;

    // Cruzamos los bodies
    for (i = _bodies.begin(); i != _bodies.end(); ++i) {
        for (j = _bodies.begin(); j != _bodies.end(); ++j) {
            // Si i != j y tenemos callback para esos tipos
            if (i != j && existsCallback((*i)->getType(), (*j)->getType(), &collisionCallback)) {
                // Filtro de esferas
                _sphereA->setCenter((*i)->getPosition());
                _sphereB->setCenter((*j)->getPosition());

                // Si pasan el filtro de esferas, llamamos al callback
                if (Shape::getCollision(_sphereA, _sphereB) && Body::getCollision(*i, *j))
                    collisionCallback(*i, *j);
            }
        }
    }
}



bool CollisionManager::existsCallback(int typeA, int typeB, CollisionCallback* collisionCallback) {
    // Buscamos una entrada para typeA
    CollisionCallbackTable::iterator i = _collisionCallbackTable.find(typeA);
    
    if (i == _collisionCallbackTable.end())
        return false;

    // Bucamos una entrada en typeA para typeB
    boost::unordered_map<int, CollisionCallback>::iterator j = i->second.find(typeB);

    if (j == i->second.end())
        return false;

    // Hemos encontrado el callback
    *collisionCallback = j->second;
    return true;
}

